"""A client for the Health Assessment Workspace Collaborative (HAWC)."""

from .animal import AnimalClient
from .assessment import AssessmentClient
from .client import BaseClient
from .epi import EpiClient
from .epimeta import EpiMetaClient
from .epiv2 import EpiV2Client
from .exceptions import HawcClientException, HawcServerException
from .interactive import InteractiveHawcClient
from .invitro import InvitroClient
from .literature import LiteratureClient
from .riskofbias import RiskOfBiasClient
from .session import HawcSession
from .study import StudyClient
from .summary import SummaryClient
from .vocab import VocabClient

__version__ = "2024.4"
__all__ = ["BaseClient", "HawcClient", "HawcClientException", "HawcServerException"]


class HawcClient(BaseClient):
    """
    HAWC Client.

    Usage:
        client = HawcClient("https://hawcproject.org")
        # If authentication is needed...
        client.authenticate("username","password")
        # To make requests...
        client.<namespace>.<method>()

    Below are the available namespaces:
        animal, epi, epimeta, invitro, lit, riskofbias, summary
    """

    def __init__(self, root_url: str = "https://hawcproject.org"):
        super().__init__(HawcSession(root_url))

        self.animal = AnimalClient(self.session)
        self.assessment = AssessmentClient(self.session)
        self.epi = EpiClient(self.session)
        self.epimeta = EpiMetaClient(self.session)
        self.epiv2 = EpiV2Client(self.session)
        self.invitro = InvitroClient(self.session)
        self.lit = LiteratureClient(self.session)
        self.riskofbias = RiskOfBiasClient(self.session)
        self.summary = SummaryClient(self.session)
        self.study = StudyClient(self.session)
        self.vocab = VocabClient(self.session)

    def authenticate(self, email: str, password: str):
        """
        Authenticate a user session

        Args:
            email (str): email to authenticate
            password (str): password to authenticate
        """
        self.session.authenticate(email, password)

    def set_authentication_token(self, token: str, login: bool = False) -> bool:
        """
        Set authentication token for hawc client session.

        Args:
            token (str): authentication token from your user profile
            login (bool, default False): if True, creates a django cookie-based session for HAWC
                similar to a standard web-browser. If False (default), creates a token-
                based django session for using the API. Requests using cookie-based sessions
                require CSRF tokens, so form functionality will be limited.

        Returns
            bool: Returns true if session is valid
        """
        return self.session.set_authentication_token(token, login)

    def interactive(self, headless: bool = True, timeout: float = 60) -> InteractiveHawcClient:
        """Generate an interactive client to download images from HAWC.

        This returns a async context manager of the interactive client; you'll also need to
        set the login=True method when authenticating to start a web-session.

        Args:
            headless (bool, default True): Run browser in headless mode.
            timeout (float, default 60): Maximum page timeout, in seconds.

        Example:

            client = HawcClient("https://hawc.com")
            client.set_authentication_token(my_api_token, login=True)
            async with client.interactive() as iclient:
                await iclient.download_visual(123456, "filename.png")

        """
        return InteractiveHawcClient(client=self, headless=headless, timeout=timeout)



import pandas as pd

from .client import BaseClient


class EpiMetaClient(BaseClient):
    """
    Client class for epidemiology metadata requests.
    """

    def data(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves epidemiology metadata for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Epidemiology metadata
        """
        url = f"{self.session.root_url}/epi-meta/api/assessment/{assessment_id}/export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)



import pandas as pd

from .client import BaseClient


class StudyClient(BaseClient):
    """
    Client class for study requests.
    """

    def create(self, reference_id: int, data: dict | None = None) -> dict:
        """
        Creates a study using a given reference ID.

        Args:
            reference_id (int): Reference ID to create study from.
            data (dict, optional): Dict containing any additional field/value pairings for the study.
                Possible pairings are:
                    short_citation (str): Short study citation, can be used as identifier.
                    full_citation (str): Complete study citation.
                    bioassay: bool (study contains animal bioassay data)
                    epi: bool (study contains epidemiology data)
                    epi_meta: bool (study contains epidemiology meta-analysis data)
                    in_vitro: bool (study contains in-vitro data)
                    coi_reported: int (COI reported, see
                        hawc.apps.study.constants.CoiReported for choices)
                    coi_details: str (COI details; use the COI declaration when available)
                    funding_source: str (any funding source information)
                    study_identifier: str (reference descriptor for assessment-tracking,
                        for example, "{Author, year, #EndNoteNumber}")
                    contact_author: bool (was the author contacted for clarification/additional data)
                    ask_author: str (correspondence details)
                    published: bool (study, evaluation, and extraction details may be visible to permitted users)
                    summary: str (often left blank, used to add comments on data extration)
                    editable: bool (project-managers/team-members allowed to edit this study)
                Defaults to {empty}.

        Returns:
            dict: JSON of the created study
        """
        if data is None:
            data = {}
        data["reference_id"] = reference_id

        url = f"{self.session.root_url}/study/api/study/"
        return self.session.post(url, data).json()

    def studies(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all studies for the given assessment.
        Args:
            assessment_id (int): Assessment ID
        Returns:
            pd.DataFrame: A dataframe for a study
        """
        url = f"{self.session.root_url}/study/api/study/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def create_from_identifier(
        self, db_type: str, db_id: str | int, assessment_id: int, **kwargs
    ) -> dict:
        """
        Creates a study using the given identifier.
        The identifier and reference are also created as needed.

        Args:
            db_type (str): Database type (HERO, PUBMED, DOI)
            db_id (Union[str,int]): Database ID
            assessment_id (int): Assessment ID to create the study under

        Returns:
            dict: JSON of the created study
        """
        payload = {"db_type": db_type, "db_id": db_id, "assessment_id": assessment_id, **kwargs}
        url = f"{self.session.root_url}/study/api/study/create-from-identifier/"
        return self.session.post(url, payload).json()



import pandas as pd

from .client import BaseClient


class SummaryClient(BaseClient):
    """
    Client class for summary requests.
    """

    def create_visual(self, data: dict) -> dict:
        """Create a new visual

        Args:
            data (dict): Required metadata for object creation.
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - visual_type (int): Constant representing visual type
                - evidence_type (int): Constant representing evidence type
                - published (bool): visual is published for public view
                - settings (dict): object settings (must be valid JSON)
                - assessment (int): assessment ID
                - prefilters (dict): object prefilters (must be valid JSON)
                - caption (str): Visual caption

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/summary/api/visual/"
        return self.session.post(url, data=data).json()

    def update_visual(self, visual_id: int, data: dict) -> dict:
        """Create a new visual

        Args:
            id (int): Visual identifier
            data (dict): Metadata to update
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - visual_type (int): Constant representing visual type
                - evidence_type (int): Constant representing evidence type
                - published (bool): visual is published for public view
                - settings (dict): object settings (must be valid JSON)
                - assessment (int): assessment ID
                - prefilters (dict): object prefilters (must be valid JSON)
                - caption (str): Visual caption

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        return self.session.patch(url, data=data).json()

    def delete_visual(self, visual_id: int):
        """Delete a visual.

        Args:
            visual_id (int): ID of the visual to delete

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        self.session.delete(url)

    def get_visual(self, visual_id: int) -> dict:
        """Get a visual.

        Args:
            visual_id (int): ID of the visual to read

        Returns:
            Response: A response object, which contains the visual if successful.
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        return self.session.get(url).json()

    def visual_list(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves a visual list for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Visual list
        """
        url = f"{self.session.root_url}/summary/api/visual/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def create_datapivot(self, data: dict) -> dict:
        """Create a new data pivot (query)

        Args:
            data (dict): Required metadata for object creation.
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - evidence_type (int): Constant representing type of evidence used in data pivot
                     (see hawc.apps.study.constants.StudyType)
                - export_style (int): Constant representing how the level at which data are aggregated,
                     and therefore which columns and types of data are presented in the export, for use
                     in the visual (see hawc.apps.summary.constants.ExportStyle)
                - preferred_units: List of preferred dose-values IDs, in order of preference.
                     If empty, dose-units will be random for each endpoint
                     presented. This setting may used for comparing
                     percent-response, where dose-units are not needed, or for
                     creating one plot similar, but not identical, dose-units.
                - published (bool): datapivot is published for public view
                - settings (str): JSON of object settings
                - assessment (int): assessment ID
                - prefilters (str): JSON of object prefilters
                - caption (str): Data pivot caption

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/summary/api/data_pivot_query/"
        return self.session.post(url, data=data).json()

    def update_datapivot(self, datapivot_id: int, data: dict) -> dict:
        """Update an existing data pivot (query)

        Args:
            id (int): Data pivot identifier
            data (dict): Required metadata for object creation.
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - evidence_type (int): Constant representing type of evidence used in data pivot
                     (see hawc.apps.study.constants.StudyType)
                - export_style (int): Constant representing how the level at which data are aggregated,
                     and therefore which columns and types of data are presented in the export, for use
                     in the visual (see hawc.apps.summary.constants.ExportStyle)
                - preferred_units: List of preferred dose-values IDs, in order of preference.
                     If empty, dose-units will be random for each endpoint
                     presented. This setting may used for comparing
                     percent-response, where dose-units are not needed, or for
                     creating one plot similar, but not identical, dose-units.
                - published (bool): datapivot is published for public view
                - settings (str): JSON of object settings
                - assessment (int): assessment ID
                - prefilters (str): JSON of object prefilters
                - caption (str): Data pivot caption

        Returns:
            dict: The resulting object, if update was successful
        """
        url = f"{self.session.root_url}/summary/api/data_pivot_query/{datapivot_id}/"
        return self.session.patch(url, data=data).json()

    def get_datapivot(self, datapivot_id: int):
        """Get a data pivot (query).

        Args:
            visual_id (int): ID of the visual to read

        Returns:
            Response: A response object, which contains the visual if successful.

        """
        url = f"{self.session.root_url}/summary/api/data_pivot_query/{datapivot_id}/"
        return self.session.get(url).json()

    def delete_datapivot(self, datapivot_id: int):
        """Delete a data pivot (query).

        Args:
            visual_id (int): ID of the visual to delete

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        url = f"{self.session.root_url}/summary/api/data_pivot_query/{datapivot_id}/"
        self.session.delete(url)

    def datapivot_list(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves a data pivot list for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Data Pivot list
        """
        url = f"{self.session.root_url}/summary/api/data_pivot/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def table_list(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves a table list for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Data Pivot list
        """
        url = f"{self.session.root_url}/summary/api/summary-table/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)



import pandas as pd
from rapidfuzz import process


def fuzz_match(
    src_df: pd.DataFrame,
    dst_df: pd.DataFrame,
    src_match_column: str,
    dst_match_column: str,
    dst_key_column: str,
    match_prefix="matched",
) -> pd.DataFrame:
    """Returns a copy of the source dataframe with a matching key value in destination.

    Args:
        src_df (pd.DataFrame): The source dataframe
        dst_df (pd.DataFrame): The matching dataframe
        src_match_column (str): The column name to attempt to match in the source
        dst_match_column (str): The column name to match to in the destination
        dst_key_column (str): The key value to return for a match
        match_prefix (str, optional): Prefix for new columns; to "matched".

    Returns:
        pd.DataFrame: _description_
    """
    comparitor = dst_df[dst_match_column].values

    def func(row):
        # returns (text, score, index)
        return process.extractOne(row[src_match_column], comparitor)

    applied_df = src_df.apply(func, axis="columns", result_type="expand")
    applied_df.loc[:, 3] = dst_df.iloc[applied_df[2].values][dst_key_column].values
    applied_df = applied_df[[3, 1, 0]].rename(
        columns={3: f"{match_prefix}_key", 1: f"{match_prefix}_score", 0: f"{match_prefix}_text"}
    )
    return pd.concat([src_df, applied_df], axis="columns")



from typing import Any


class HawcException(Exception):
    def __init__(self, status_code: int, message: Any):
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return f"<{self.status_code}> {self.message}"


class HawcClientException(HawcException):
    """An exception occurred in the HAWC client module."""


class HawcServerException(HawcException):
    """An exception occurred on the HAWC server."""


class ContentUnavailable(HawcClientException):
    """Expected content is not available."""



import pandas as pd

from .client import BaseClient
from .utils import fuzz_match


class RiskOfBiasClient(BaseClient):
    """
    Client class for risk of bias requests.
    """

    def export(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves risk of bias data for the given assessment.
        This includes only final reviews.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Risk of bias data
        """
        url = f"{self.session.root_url}/rob/api/assessment/{assessment_id}/export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def full_export(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves full risk of bias data for the given assessment.
        This includes user-level reviews.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Full risk of bias data
        """
        url = f"{self.session.root_url}/rob/api/assessment/{assessment_id}/full-export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def create(
        self, study_id: int, author_id: int, active: bool, final: bool, scores: list[dict]
    ) -> dict:
        """
        Create a Risk of Bias review for a study. The review must be complete and contain answers for
        all required metrics.

        Args:
            study_id (int): id of study.
            author_id (int): id of author of the Risk of Bias data.
            active (bool): create the new Risk of Bias data as active or not
            final (bool): create the new Risk of Bias data as final or not
            scores (list[dict]): List of scores. Each element of the list is a dict containing the following
                                 string keys / expected values:
                * "metric_id" (int): the id of the metric for this score
                * "is_default" (bool): create this score as default or not
                * "label" (str): label for this score
                * "notes" (str): notes for this core
                * "score" (int): numeric score value. Actual legal values for this are dependent on the value
                                 of the HAWC_FLAVOR setting for this instance of HAWC and correspond to readable
                                 values like "Critically deficient" or "++". See also:
                                    hawc.apps.riskofbias.constants.SCORE_CHOICES
                                    + hawc.apps.riskofbias.constants.SCORE_SYMBOLS
                * bias_direction (int, optional): bias direction
                * "overridden_objects" (list[dict], optional): a list of overrides for this particular score. Optional.
                                                     Each element of this list is a dict containing the
                                                     following string keys / expected values:
                    * "content_type_name" (str): the name of the data type relevant to this override.
                    * "object_id" (int): the id of the particular instance of that data type relevant to this override.
        """
        payload = {
            "study_id": study_id,
            "author_id": author_id,
            "active": active,
            "final": final,
            "scores": scores,
        }
        url = f"{self.session.root_url}/rob/api/review/"
        return self.session.post(url, payload).json()

    def bulk_rob_copy(
        self,
        src_assessment_id: int,
        dst_assessment_id: int,
        src_dst_study_ids: list[tuple[int, int]],
        src_dst_metric_ids: list[tuple[int, int]],
        copy_mode: int,
        author_mode: int,
        dst_author_id: int | None = None,
    ):
        """
        Copy final scores from a subset of studies from one assessment as the scores in a
        different assessment. Useful when an assessment is cloned or repurposed and existing
        evaluations should be used in a new evaluation. You must be a project-manager on both
        assessments.

        Args:
            src_assessment_id (int): source assessment
            dst_assessment_id (int): destination assessment
            src_dst_study_ids (list[tuple[int, int]]): source study id, destination study id pairings
            src_dst_metric_ids (list[tuple[int, int]]): source metric id, destination metric id pairings
            copy_mode (int): enum for copy mode
                1 = src active riskofbias -> dest active risk of bias
                2 = src final riskofbias -> dest initial risk of bias
            author_mode (int): enum for author mode
                1: original authors are preserved
                2: authors are overwritten by given dst_author_id
            dst_author_id (Optional[int]): author for destination RoBs when author_mode = 2.

        Returns:
            dict: Log information and mapping of all source ids to destination ids
        """

        payload = {
            "src_assessment_id": src_assessment_id,
            "dst_assessment_id": dst_assessment_id,
            "src_dst_study_ids": src_dst_study_ids,
            "src_dst_metric_ids": src_dst_metric_ids,
            "copy_mode": copy_mode,
            "author_mode": author_mode,
        }
        if dst_author_id is not None:
            payload["dst_author_id"] = dst_author_id

        url = f"{self.session.root_url}/rob/api/assessment/bulk_rob_copy/"
        return self.session.post(url, payload).json()

    def metrics(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all metrics for the given assessment.
        Args:
            assessment_id (int): Assessment ID
        Returns:
            pd.DataFrame: A dataframe of metrics
        """
        url = f"{self.session.root_url}/rob/api/metric/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def reviews(self, assessment_id: int, study_id: int | None = None) -> dict:
        """
        Retrieves all reviews for the given assessment. Must be team member or higher
        Args:
            assessment_id (int): Assessment ID
            study_id (int | None): A study ID; returns results for a single study
        Returns:
            dict: A dictionary of reviews
        """
        url = f"{self.session.root_url}/rob/api/review/?assessment_id={assessment_id}"
        if study_id:
            url += f"&study={study_id}"
        return self.session.get(url).json()

    def final_reviews(self, assessment_id: int, study_id: int | None = None) -> dict:
        """
        Retrieves all final reviews for the given assessment.
        Args:
            assessment_id (int): Assessment ID
            study_id (int | None): A study ID; returns results for a single study
        Returns:
            dict: A dictionary of final reviews
        """
        url = f"{self.session.root_url}/rob/api/review/final/?assessment_id={assessment_id}"
        if study_id:
            url += f"&study={study_id}"
        return self.session.get(url).json()

    def compare_metrics(
        self, src_assessment_id: int, dest_assessment_id: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Match metrics across two different assessments, returning best-matched results.

        This uses a fuzzy-text matching algorithm to attempt to match text between two different
        assessments.  It returns the best metric-match from the destination, for each metric in the
        source.

        Adds three columns to the source metric dataframe:
        - fuzz_key: the matching metric id form the destination
        - fuzz_score: a similarity score between 0 and 100. 100 is an exact match
        - fuzz_text: the text from the destination metric which was compared

        Args:
            src_assessment_id (int): The source assessment, try to match each metric
            dest_assessment_id (int): The target assessment, where matches are returned from

        Returns:
            tuple[pd.DataFrame, pd.DataFrame]: Metrics with matches in source, Metrics in destination
        """
        src = self.metrics(src_assessment_id).assign(assessment_id=src_assessment_id)
        dest = self.metrics(dest_assessment_id).assign(assessment_id=dest_assessment_id)
        src.loc[:, "_comparison"] = src.short_name + " " + src.name + " " + src.description
        dest.loc[:, "_comparison"] = dest.short_name + " " + dest.name + " " + dest.description
        matched = fuzz_match(src, dest, "_comparison", "_comparison", "id").drop(
            columns=["_comparison"]
        )
        return matched, dest



import pandas as pd
from requests import Response

from .client import BaseClient


class LiteratureClient(BaseClient):
    """
    Client class for literature requests.
    """

    def _import(
        self, source: int, assessment_id: int, title: str, description: str, ids: list[int]
    ) -> dict:
        """
        Imports a list of IDs as literature references for the given assessment.

        Args:
            source (int): Database source (1 for PubMed, 2 for HERO)
            assessment_id (int): Assessment ID
            title (str): Title of import
            description (str): Description of import
            ids (list[int]): IDs

        Returns:
            dict: JSON response
        """
        payload = {
            "assessment": assessment_id,
            "search_type": "i",
            "source": source,
            "title": title,
            "description": description,
            "search_string": ",".join(str(id_) for id_ in ids),
        }
        url = f"{self.session.root_url}/lit/api/search/"
        return self.session.post(url, payload).json()

    def import_hero(self, assessment_id: int, title: str, description: str, ids: list[int]) -> dict:
        """
        Imports a list of HERO IDs as literature references for the given assessment.

        Args:
            assessment_id (int): Assessment ID
            title (str): Title of import
            description (str): Description of import
            ids (list[int]): HERO IDs

        Returns:
            dict: JSON response
        """
        return self._import(2, assessment_id, title, description, ids)

    def import_pubmed(
        self, assessment_id: int, title: str, description: str, ids: list[int]
    ) -> dict:
        """
        Imports a list of PubMed IDs as literature references for the given assessment.

        Args:
            assessment_id (int): Assessment ID
            title (str): Title of import
            description (str): Description of import
            ids (list[int]): PubMed IDs

        Returns:
            dict: JSON response
        """
        return self._import(1, assessment_id, title, description, ids)

    def tags(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all of the tags for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Assessment tags
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/tags/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def get_tagtree(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves the nested tag tree for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            dict: JSON representation of the tag tree
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/tagtree/"
        response_json = self.session.get(url).json()
        return response_json["tree"]

    def clone_tagtree(self, source_assessment_id: int, target_assessment_id: int) -> pd.DataFrame:
        """
        Copies the tag tree from one assessment to another.

        Args:
            source_assessment_id (int): Assessment ID to copy tag tree from
            target_assessment_id (int): Assessment ID to copy tag tree to

        Returns:
            dict: JSON representation of the new tag tree
        """
        fetch_url = f"{self.session.root_url}/lit/api/assessment/{source_assessment_id}/tagtree/"
        tree = self.session.get(fetch_url).json()

        update_url = f"{self.session.root_url}/lit/api/assessment/{target_assessment_id}/tagtree/"
        update_response_json = self.session.post(update_url, tree).json()

        return update_response_json["tree"]

    def update_tagtree(self, assessment_id: int, tags: list[dict]) -> pd.DataFrame:
        """
        Updates the tag tree.

        Args:
            assessment_id (int): Assessment ID to update
            tags (list[dict]): tag definitions. For each tag dict element, "name" is required. "slug" is
               optional. "children" is optional and should contain a recursive list containing valid tags.

        Returns:
            dict: JSON representation of the new tag tree. If errors, a JSON list containing details.
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/tagtree/"
        response_json = self.session.post(url, {"tree": tags}).json()
        return response_json["tree"]

    def reference_tags(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves the literature references and their corresponding tags for a given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: References with corresponding tags
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/reference-tags/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def import_reference_tags(
        self, assessment_id: int, csv: str, operation: str = "append", dry_run: bool = False
    ) -> pd.DataFrame:
        """
        Imports a CSV of reference IDs with corresponding tag IDs to the given assessment.

        Args:
            assessment_id (int): Assessment ID
            csv (str): Reference IDs to tag ID mapping. The header of this CSV string should be "reference_id,tag_id".
            operation (str, optional): Either add new references tags to existing (`append`), or replace current tag mappings (`replace`). Defaults to "append".
            dry_run (bool, optional): If set to True, runs validation checks but does not execute

        Returns:
            pd.DataFrame: All tag mappings for the selected `assessment_id`, after the requested changes
        """
        payload = {"csv": csv, "operation": operation, "dry_run": dry_run}
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/reference-tags/"
        response_json = self.session.post(url, payload).json()
        return pd.DataFrame(response_json)

    def reference_ids(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves the reference IDs and corresponding HERO/PubMed/etc IDs for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Reference IDs and HERO/PubMed/etc IDs
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/reference-ids/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def references(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all references for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: References data
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/reference-export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def reference_user_tags(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all user tag for all references for an assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: References user tag data
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/user-tag-export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def reference(self, reference_id: int) -> dict:
        """
        Retrieves the selected reference.

        Args:
            reference_id (int): ID of the reference to retrieve

        Returns:
            dict: JSON representation of the reference
        """
        url = f"{self.session.root_url}/lit/api/reference/{reference_id}/"
        response_json = self.session.get(url).json()
        return response_json

    def update_reference(self, reference_id: int, **kwargs) -> dict:
        """
        Updates reference with given values. Fields not passed as parameters
        are unchanged.

        Args:
            reference_id (int): ID of reference to update
            **kwargs (optional): Named parameters of fields to update in reference. Example parameters:
                title (str): title of the reference
                abstract (str): reference abstract
                tags (list[int]): tag IDs to apply to reference;
                    replaces the existing tags

        Example Usage:
            updated_reference_json = client.lit.update_reference(
                reference_id = 1,
                title = "reference",
                tags = [1,2,3]
            )

        Returns:
            dict: JSON representation of the updated reference.
        """
        url = f"{self.session.root_url}/lit/api/reference/{reference_id}/"
        response_json = self.session.patch(url, kwargs).json()
        return response_json

    def delete_reference(self, reference_id: int) -> None:
        """
        Deletes the selected reference. This also removes the reference from any
        searches/imports which may have included the reference. If data was
        extracted with this reference and it is associated with bioassay or epi
        extractions they will also be removed.

        Args:
            reference_id (int): ID of reference to delete

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        url = f"{self.session.root_url}/lit/api/reference/{reference_id}/"
        self.session.delete(url)

    def replace_hero(self, assessment_id: int, replace: list[list[int]]) -> None:
        """
        Replace HERO ID associated with each reference with a new HERO ID. Reference
        fields are updated using the new HERO ID's reference metadata.  This request is
        throttled; can only be executed once per minute.

        This method schedules a task to be executed when workers are available; task completion
        therefore is not guaranteed even with a successful response.

        Args:
            assessment_id (int): Assessment ID for all references in the list.
            replace (list[list[int]]): List of reference ID / new HERO ID pairings, both values
                should be integers, ex., [[reference_id, hero_id], ... ]

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        body = {"replace": replace}
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/replace-hero/"
        self.session.post(url, body)

    def update_references_from_hero(self, assessment_id: int) -> None:
        """
        Updates the fields of all HERO references in an assessment with the most recent metadata
        from HERO. This request is throttled; can only be executed once per minute.

        This method schedules a task to be executed when workers are available; task completion
        therefore is not guaranteed even with a successful response.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        url = f"{self.session.root_url}/lit/api/assessment/{assessment_id}/update-reference-metadata-from-hero/"
        self.session.post(url)

    def create_tag(self, data: dict) -> dict:
        """
        Create a tag.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful.
        """
        url = f"{self.session.root_url}/lit/api/tags/"
        response = self.session.post(url, data)
        return response.json()

    def update_tag(self, tag_id: int, data: dict) -> dict:
        """
        Update an existing tag.

        Args:
            tag_id (int): Tag ID
            data (dict): fields to update in tag

        Returns:
            dict: The resulting object, if update was successful.
        """
        url = f"{self.session.root_url}/lit/api/tags/{tag_id}/"
        response = self.session.patch(url, data)
        return response.json()

    def delete_tag(self, tag_id: int) -> Response:
        """
        Delete a tag.

        Args:
            tag_id (int): Tag ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/lit/api/tags/{tag_id}/"
        return self.session.delete(url)

    def move_tag(self, tag_id: int, new_index: int) -> dict:
        """
        Change the order of a tag within its tagtree.

        Args:
            tag_id (int): Tag ID
            new_index: Index to place this tag in relation to its siblings in the tagtree

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/lit/api/tags/{tag_id}/move/"
        body = {"newIndex": new_index}
        return self.session.patch(url, body)
